==> kite-1.0b3.tar.gz: kite-1.0b3/vm/kite_thread.c <==
/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>
#include <setjmp.h>
#include <pthread.h>
#include <signal.h>
#include "objs/kite_object.h"
#include "kite_vm.h"

struct kite_object_t *kite_signal_handlers[NSIG + 1];
static kite_thread_t *signal_handler_thread = NULL;

/*****************************************************************************
 * Interface to user-defined signal handlers
 ****************************************************************************/
void kite_handle_signal(int sig)
{
    kite_thread_t *thd = signal_handler_thread;

    if (kite_signal_handlers[sig] && thd) {
        kite_object_t *list = kite_new_list(thd);
        list->builtin_data.listvalue.car = kite_new_integer(thd, sig);
        kite_vm_call_object(thd, kite_new_null(thd), 
            kite_signal_handlers[sig], list);
        kite_dereference_object(list);
    }
}

/*****************************************************************************
 * Create a new thread (without starting)
 ****************************************************************************/
kite_thread_t *kite_new_thread_without_start(kite_vm_t *vm)
{
    volatile kite_thread_t *thd;
    kite_thread_t *exchndl;
    assert(vm != NULL);

    thd = malloc(sizeof(kite_thread_t));
    if (!thd) return NULL;
    
    thd->vm = vm;
    thd->running_stack = NULL;
    thd->func_stack = NULL;
    thd->start = NULL;
    thd->exception = NULL;

    /* Add to end of thread list */
    vm->numthreads++;
    vm->thread_list = realloc(vm->thread_list,
        sizeof(kite_thread_t*)*vm->numthreads);
    thd->entry = &vm->thread_list[vm->numthreads - 1];
    vm->thread_list[vm->numthreads - 1] = (kite_thread_t*)thd;

    /* Set up signals */
    if (!signal_handler_thread) {
        signal_handler_thread = thd;
    }

    return (kite_thread_t*)thd;
}

/*****************************************************************************
 * Perform thread cleanup
 ****************************************************************************/
static void kite_cleanup_thread(kite_thread_t *thd)
{
    kite_vm_t *vm = thd->vm;
    size_t bytes = 
      vm->thread_list + sizeof(kite_thread_t*) * (vm->numthreads - 1)
      - (thd->entry + 1);

    if (vm->numthreads == 1) {
        free(vm->thread_list);
        vm->thread_list = NULL;
        vm->numthreads = 0;
    } else {
        memmove(thd->entry, thd->entry + 1, bytes);
        vm->numthreads--;
        vm->thread_list = 
            realloc(vm->thread_list, sizeof(kite_thread_t*) * vm->numthreads);
    }
    
    while(thd->running_stack && thd->running_stack->length > 0) {
        kite_object_t *obj = kite_pop_stack(&thd->running_stack);
        kite_dereference_object(obj);
    }
    if (thd->running_stack) kite_destroy_stack(&thd->running_stack);

    while(thd->func_stack && thd->func_stack->length) {
        kite_funccall_info_t *fs = kite_pop_stack(&thd->func_stack);
        if (fs->locals) kite_destruct_symtab(thd, &fs->locals, TRUE);
    }
    if (thd->func_stack) kite_destroy_stack(&thd->func_stack);

    if (signal_handler_thread == thd) {
        if (vm->numthreads) signal_handler_thread = vm->thread_list[0];
        else signal_handler_thread = NULL;
    }

    if (vm->thread_list == NULL)
        kite_gc_destroy_all(thd);
    else
    {
        kite_list_t *lst = vm->gc_begin;
        while(lst)
        {
            kite_object_t *obj = lst->obj;
            obj->owner_thread = vm->thread_list[0];
            lst = lst->next;
        }
    }
    
    free(thd);
}

/* 
 * 4MB is enough to ensure we throw an exception before we overrun the stack
 * pthreads gave us.
 */
#define PTHREAD_STACK_SIZE 4194304

/*****************************************************************************
 * Start a new thread (using bytecode)
 ****************************************************************************/
void kite_start_bytecode(kite_thread_t *thd, void *start)
{
    pthread_attr_t attr;
    sigset_t mask;
    pthread_attr_init(&attr);
    pthread_attr_setstacksize(&attr, PTHREAD_STACK_SIZE);
    
#ifdef __APPLE__
    sigfillset(&mask); /* Unmask all allowed signals */
    pthread_sigmask(SIG_UNBLOCK, &mask, NULL);
#endif // __APPLE__

    thd->start = start;
    pthread_create(&thd->thread, &attr, kite_vm_execute, thd);
    pthread_attr_destroy(&attr);
}

/*****************************************************************************
 * Start a new thread (using compiled)
 ****************************************************************************/
void kite_start_compiled(kite_thread_t *thd, void *start)
{
    pthread_attr_t attr;
    pthread_attr_init(&attr);
    pthread_attr_setstacksize(&attr, PTHREAD_STACK_SIZE);
    
    thd->start = NULL;
    pthread_create(&thd->thread, &attr, start, thd);
    pthread_attr_destroy(&attr);
}

/*****************************************************************************
 * Create a new thread (using bytecode)
 ****************************************************************************/
kite_thread_t *kite_new_thread_bytecode(kite_vm_t *vm, void *start)
{
    kite_thread_t *thd = kite_new_thread_without_start(vm);
    assert(start != NULL);

    kite_start_bytecode(thd, start);
    return thd;
}

/*****************************************************************************
 * Create a new thread (using compiled code)
 ****************************************************************************/
kite_thread_t *kite_new_thread_compiled(kite_vm_t *vm, void *start)
{
    kite_thread_t *thd = kite_new_thread_without_start(vm);
    assert(start != NULL);
    
    kite_start_compiled(thd, start);
    return thd;
}

/*****************************************************************************
 * Wait for given thread to exit
 ****************************************************************************/
void kite_join_thread(kite_vm_t *vm, kite_thread_t *thd)
{
    assert(thd->vm == vm);
    if (thd->start) {
        pthread_join(thd->thread, NULL);
        pthread_detach(thd->thread);
    }

    /* Remove thread from thread list */
    kite_cleanup_thread(thd);
}

/*****************************************************************************
 * Exit given thread
 ****************************************************************************/
void kite_exit_thread(kite_vm_t *vm, kite_thread_t *thd)
{
    assert(thd->vm == vm);
    if (thd->start) pthread_cancel(thd->thread);
    kite_join_thread(vm, thd);
}

==> kite_1.0.2.orig.tar.gz: kite-1.0.2/vm/kite_thread.c <==
/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <setjmp.h>
#ifndef WIN32
#include <pthread.h>
#endif /* WIN32 */
#include <signal.h>
#include "objs/kite_object.h"
#include "objs/kite_gc.h"
#include "kite_vm.h"
#ifndef WIN32
#include <unistd.h>
#endif /* WIN32 */

struct kite_object_t *kite_signal_handlers[NSIG + 1];
static kite_thread_t *signal_handler_thread = NULL;

/*****************************************************************************
 * Interface to user-defined signal handlers
 ****************************************************************************/
void kite_handle_signal(int sig)
{
    kite_thread_t *thd = signal_handler_thread;

    if (kite_signal_handlers[sig] && thd) {
        kite_object_t *list = kite_new_list(thd);
        list->builtin_data.listvalue.car = kite_new_integer(thd, sig);
        kite_vm_call_object(thd, kite_new_null(thd), 
            kite_signal_handlers[sig], list);
        kite_dereference_object(list);
    }
}

/*****************************************************************************
 * Create a new thread (without starting)
 ****************************************************************************/
kite_thread_t *kite_new_thread_without_start(kite_vm_t *vm)
{
    kite_thread_t *thd;
    assert(vm != NULL);

#ifndef HAVE_GC_H
    thd = (kite_thread_t*)calloc(1, sizeof(kite_thread_t));
#else
    thd = (kite_thread_t*)GC_malloc(sizeof(kite_thread_t));
#endif /* HAVE_GC_H */
    if (!thd) return NULL;
    
    thd->vm = vm;
    thd->running_stack = NULL;
    thd->func_stack = NULL;
    thd->start = NULL;
    thd->exception = NULL;

    /* Add to end of thread list */
    vm->numthreads++;
#ifndef HAVE_GC_H
    vm->thread_list = (kite_thread_t**)realloc(vm->thread_list,
        sizeof(kite_thread_t*)*vm->numthreads);
#else
    vm->thread_list = (kite_thread_t**)GC_realloc(vm->thread_list, sizeof(kite_thread_t)*vm->numthreads);
#endif /* HAVE_GC_H */
    thd->entry = &vm->thread_list[vm->numthreads - 1];
    vm->thread_list[vm->numthreads - 1] = (kite_thread_t*)thd;

    /* Set up signals */
    if (!signal_handler_thread) {
        signal_handler_thread = thd;
    }

    return (kite_thread_t*)thd;
}

/*****************************************************************************
 * Perform thread cleanup
 ****************************************************************************/
static void kite_cleanup_thread(kite_thread_t *thd)
{
    kite_vm_t *vm = thd->vm;
    size_t bytes = 
      vm->thread_list + sizeof(kite_thread_t*) * (vm->numthreads - 1)
      - (thd->entry + 1);

    if (vm->numthreads == 1) {
#ifndef HAVE_GC_H
        free(vm->thread_list);
#endif /* HAVE_GC_H */
        vm->thread_list = NULL;
        vm->numthreads = 0;
    } else {
        memmove(thd->entry, thd->entry + 1, bytes);
        vm->numthreads--;
#ifndef HAVE_GC_H
        vm->thread_list = 
            (kite_thread_t**)realloc(vm->thread_list, sizeof(kite_thread_t*) * vm->numthreads);
#else
        vm->thread_list = 
            (kite_thread_t**)GC_realloc(vm->thread_list, sizeof(kite_thread_t*) * vm->numthreads);
#endif /* HAVE_GC_H */
    }
    
    while(thd->running_stack && thd->running_stack->length > 0) {
        kite_object_t *obj = kite_pop_stack(&thd->running_stack);
        kite_dereference_object(obj);
    }
    if (thd->running_stack) kite_destroy_stack(&thd->running_stack);

    while(thd->func_stack && thd->func_stack->length) {
        kite_funccall_info_t *fs = kite_pop_stack(&thd->func_stack);
        if (fs->locals) kite_destruct_symtab(thd, &fs->locals, TRUE);
    }
    if (thd->func_stack) kite_destroy_stack(&thd->func_stack);

    if (signal_handler_thread == thd) {
        if (vm->numthreads) signal_handler_thread = vm->thread_list[0];
        else signal_handler_thread = NULL;
    }

    if (vm->thread_list == NULL)
        kite_gc_destroy_all(thd);
    else
    {
#ifndef HAVE_GC_H
        kite_list_t *lst = vm->gc_begin;
        while(lst)
        {
            kite_object_t *obj = lst->obj;
            obj->owner_thread = vm->thread_list[0];
            lst = lst->next;
        }
#endif /* HAVE_GC_H */
    }

#ifndef HAVE_GC_H
    free(thd);
#endif /* HAVE_GC_H */
}

/* 
 * 4MB is enough to ensure we throw an exception before we overrun the stack
 * pthreads gave us.
 */
#define PTHREAD_STACK_SIZE 4194304

/*****************************************************************************
 * Start a new thread (using bytecode)
 ****************************************************************************/
#ifndef WIN32
void kite_start_bytecode(kite_thread_t *thd, void *start)
{
    pthread_attr_t attr;

#ifdef __APPLE__
    sigset_t mask;
#endif /* __APPLE__ */

    if (thd->vm->thread_list[0] == thd) 
    {
        thd->start = start;
        kite_vm_execute(thd);
        return;
    }
    
    pthread_attr_init(&attr);
    pthread_attr_setstacksize(&attr, PTHREAD_STACK_SIZE);
    
#ifdef __APPLE__
    sigfillset(&mask); /* Unmask all allowed signals */
    pthread_sigmask(SIG_UNBLOCK, &mask, NULL);
#endif // __APPLE__

    thd->start = start;
    pthread_create(&thd->thread, &attr, kite_vm_execute, thd);
    pthread_attr_destroy(&attr);
}
#else
void kite_start_bytecode(kite_thread_t *thd, void *start)
{
    if (thd->vm->thread_list[0] == thd) 
    {
        thd->start = start;
        kite_vm_execute(thd);
        return;
    }

    thd->start = start;
    thd->thread = CreateThread(NULL, PTHREAD_STACK_SIZE, kite_vm_execute, thd, 0, NULL);
}
#endif /* WIN32 */

/*****************************************************************************
 * Start a new thread (using compiled)
 ****************************************************************************/
#ifndef WIN32
void kite_start_compiled(kite_thread_t *thd, void *start)
{
    pthread_attr_t attr;

    if (thd->vm->thread_list[0] == thd) 
    {
        void (*ptr)(kite_thread_t *) = start;
        ptr(thd);
        return;
    }
    
    pthread_attr_init(&attr);
    pthread_attr_setstacksize(&attr, PTHREAD_STACK_SIZE);
    
    thd->start = NULL;
    pthread_create(&thd->thread, &attr, start, thd);
    pthread_attr_destroy(&attr);
}
#else
void kite_start_compiled(kite_thread_t *thd, void *start)
{
    if (thd->vm->thread_list[0] == thd) 
    {
        void (*ptr)(kite_thread_t *) = start;
        ptr(thd);
        return;
    }
    
    thd->start = NULL;
    thd->thread = CreateThread(NULL, PTHREAD_STACK_SIZE, start, thd, 0, NULL);
}
#endif /* WIN32 */

/*****************************************************************************
 * Create a new thread (using bytecode)
 ****************************************************************************/
kite_thread_t *kite_new_thread_bytecode(kite_vm_t *vm, void *start)
{
    kite_thread_t *thd = kite_new_thread_without_start(vm);
    assert(start != NULL);

    kite_start_bytecode(thd, start);
    return thd;
}

/*****************************************************************************
 * Create a new thread (using compiled code)
 ****************************************************************************/
kite_thread_t *kite_new_thread_compiled(kite_vm_t *vm, void *start)
{
    kite_thread_t *thd = kite_new_thread_without_start(vm);
    assert(start != NULL);
    
    kite_start_compiled(thd, start);
    return thd;
}

/*****************************************************************************
 * Wait for given thread to exit
 ****************************************************************************/
void kite_join_thread(kite_vm_t *vm, kite_thread_t *thd)
{
    assert(thd->vm == vm);
    if (thd->vm->thread_list[0] != thd) {
#ifndef WIN32
        pthread_join(thd->thread, NULL);
        pthread_detach(thd->thread);
#else
        WaitForSingleObject(thd->thread, INFINITE);
#endif /* WIN32 */
    }

    /* Remove thread from thread list */
    kite_cleanup_thread(thd);
}

/*****************************************************************************
 * Exit given thread
 ****************************************************************************/
void kite_exit_thread(kite_vm_t *vm, kite_thread_t *thd)
{
    assert(thd->vm == vm);
#ifndef WIN32
    if (thd->start) pthread_cancel(thd->thread);
#else
    if (thd->start) TerminateThread(thd->thread, 0);
#endif /* WIN32 */
    kite_join_thread(vm, thd);
}
